#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

import types
import warnings

from functools import wraps


# TODO: Invalidate this module in favor of new 'xotl.tools.suggest' when
# implemented

# FIX: 'warnings.warn' uses in this module 'UserWarning' instead of
# 'DeprecationWarning'.  There is a way to signal the warning with the correct
# type.

DEFAULT_MSG = (
    "{funcname} is now deprecated and it will be "
    "removed{in_version}. Use {replacement} instead."
)


class DeprecationError(Exception):
    pass


# TODO: Use ``warnings.simplefilter('default', DeprecationWarning)``


# TODO: Maybe adapt all other functions in this module to use this descriptor.
# Currently, it's only being used in combination with
# xotl.tools.modules.customize to deprecate imports from xotl.tools top level
# module.
class DeprecatedImportDescriptor:
    "A descriptor that issues a deprecation warning when resolving `name`."

    def __init__(self, replacement):
        self.attr = replacement[replacement.rfind(".") + 1 :]
        self.replacement = replacement

    def __get__(self, instance, owner):
        if instance is not None:
            import warnings
            from xotl.tools.objects import import_object

            result = import_object(self.replacement)
            warnings.warn(
                "Importing {name} from xotl.tools is deprecated. "
                "You should import it from {ns}".format(
                    name=self.attr, ns=self.replacement
                ),
                UserWarning,  # DeprecationWarning is silent in ipython
            )
            return result
        else:
            return self


def _nameof(item):
    "Version of `xotl.tools.names.nameof`:func: to avoid importing it here."
    singletons = (None, True, False, Ellipsis, NotImplemented)
    res = next((str(s) for s in singletons if s is item), None)
    if res is None:
        res = ".".join([item.__module__, item.__name__])
    return res


def deprecated(
    replacement,
    msg=DEFAULT_MSG,
    deprecated_module=None,
    removed_in_version=None,
    check_version=False,
    new_name=None,
):
    """Small decorator for deprecated functions.

    Usage::

        @deprecated(new_function)
        def deprecated_function(...):
            ...

    :param replacement: Either a string or the object that replaces the
       deprecated.

    :param msg: A deprecation warning message template.  You should provide
       keyword arguments for the `format`:func: function.  Currently we pass
       the current keyword arguments: `replacement` (after some processing),
       `funcname` with the name of the currently deprecated object and
       `in_version` with the version this object is going to be removed if
       `removed_in_version` argument is not None.

       Defaults to: "{funcname} is now deprecated and it will be
       removed{in_version}.  Use {replacement} instead."

    :param removed_in_version: The version the deprecated object is going to be
       removed.

    :param check_version: If True and `removed_in_version` is not None, then
       declarations of obseleted objects will raise a DeprecationError.  This
       helps the release manager to keep the release clean.

       .. note:: Currently only works with setuptools' installed distributions.

    :param deprecated_module: If provided, the name of the module the
       deprecated object resides.  Not needed if the deprecated object is a
       function or class.

    :param new_name: If provided, it's used as the name of the
       deprecated object.  Needed to allow renaming in
       `import_deprecated`:func: helper function.

    .. note:: Deprecating some classes in Python 3 could fail.  This is
       because those classes do not declare a '__new__' par of the declared
       '__init__'.  The problem is solved if the '__new__' of the super-class
       has no arguments.  This doesn't happen in Python 2.

       To solve these cases mark the deprecation in a comment and issue the
       warning directly in the constructor code.

    .. versionchanged:: 1.4.1 Introduces removed_in_version and check_version.

    """

    def raise_if_deprecated(target, target_version):
        import pkg_resources

        pkg = _nameof(target)
        pkg, _obj = pkg.rsplit(".", 1)
        dist = None
        while not dist and pkg:
            try:
                dist = pkg_resources.get_distribution(pkg)
            except pkg_resources.DistributionNotFound:
                dist = None
                if "." in pkg:
                    pkg, _obj = pkg.rsplit(".", 1)
                else:
                    pkg, _obj = None, None  # noqa
        assert dist
        if isinstance(target_version, str):
            target_version = pkg_resources.parse_version(target_version)
        if dist.parsed_version >= target_version:
            msg = (
                "A deprecated feature %r was scheduled to be "
                "removed in version %r and it is still "
                "alive in %r!"
                % (_nameof(target), str(removed_in_version), str(dist.version))
            )
            raise DeprecationError(msg)

    def decorator(target):
        target_name = new_name if new_name else target.__name__
        if deprecated_module:
            funcname = deprecated_module + "." + target_name
        else:
            funcname = target_name
        if isinstance(replacement, (type, types.FunctionType)):
            repl_name = replacement.__module__ + "." + replacement.__name__
        else:
            repl_name = replacement
        if removed_in_version:
            in_version = " in version " + removed_in_version
        else:
            in_version = ""
        if isinstance(target, type):

            def new(*args, **kwargs):
                if check_version and removed_in_version:
                    raise_if_deprecated(target, removed_in_version)
                warnings.warn(
                    msg.format(
                        funcname=funcname, replacement=repl_name, in_version=in_version
                    ),
                    stacklevel=2,
                )
                try:
                    return target.__new__(*args, **kwargs)
                except TypeError:
                    # XXX: Some classes in Python 3 don't declare an
                    # equivalent '__new__'
                    return super(result, args[0]).__new__(args[0])

            # Code copied and adapted from xotl.tools.objects.copy_class.
            # This is done so because this module *must* not depends on any
            # other, otherwise an import cycle might be formed when
            # deprecating a class in xotl.tools.objects.
            from xotl.tools.future.types import MemberDescriptorType

            meta = type(target)
            td = target.__dict__
            iteritems = td.items
            attrs = {
                name: value
                for name, value in iteritems()
                if name
                not in ("__class__", "__mro__", "__name__", "__weakref__", "__dict__")
                # Must remove member descriptors, otherwise the old's
                # class descriptor will override those that must be
                # created here.
                if not isinstance(value, MemberDescriptorType)
            }
            attrs.update(__new__=new)
            result = meta(target_name, target.__bases__, attrs)
            return result
        else:

            @wraps(target)
            def inner(*args, **kw):
                if check_version and removed_in_version:
                    raise_if_deprecated(target, removed_in_version)
                warnings.warn(
                    msg.format(
                        funcname=funcname, replacement=repl_name, in_version=in_version
                    ),
                    stacklevel=2,
                )
                return target(*args, **kw)

            if new_name:
                inner.__name__ = new_name
            return inner

    return decorator


def import_deprecated(module, *names, **aliases):
    """Import functions deprecating them in the target module.

    The target module is the caller of this function (only intended to be
    called in the global part of a module).

    :param module: The module from which functions will be imported.  Could be
           a string, or an imported module.

    :param names: The names of the functions to import.

    :param aliases: Keys are the new names, values the old names.

    For example::

      >>> from xotl.tools.deprecation import import_deprecated
      >>> import math
      >>> import_deprecated(math, 'sin', new_cos='cos')
      >>> sin is not math.sin
      True

    Next examples are all ``True``, but them print the deprecation warning
    when executed::

      >>> sin(math.pi/2) == 1.0
      >>> new_cos(2*math.pi) == math.cos(2*math.pi)

    If no identifier is given, it is assumed equivalent as ``from module
    import *``.

    The statement ``import_deprecated('math', 'sin', new_cos='cos')`` has the
    same semantics as ``from math import sin, cos as new_cos``, but
    deprecating current module symbols.

    This function is provided for easing the deprecation of whole modules and
    should not be used to do otherwise.

    """
    from xotl.tools.future.types import func_types
    from xotl.tools.modules import force_module

    src = force_module(module)
    dst = force_module(2)
    src_name = src.__name__
    dst_name = dst.__name__
    dst = force_module(2)
    if not names and not aliases:
        # from module import *
        names = getattr(src, "__all__", None)
        if not names:
            names = (n for n in dir(src) if not n.startswith("_"))
    for name in names:
        if name not in aliases:
            aliases[name] = name
        else:
            msg = 'import_deprecated(): invalid repeated argument "{}"'
            raise ValueError(msg.format(name))
    unset = object()
    test_classes = func_types + (type,)
    for alias in aliases:
        name = aliases[alias]
        target = getattr(src, name, unset)
        if target is not unset:
            if isinstance(target, test_classes):
                replacement = src_name + "." + name
                deprecator = deprecated(
                    replacement, DEFAULT_MSG, dst_name, new_name=alias
                )
                target = deprecator(target)
            setattr(dst, alias, target)
        else:
            msg = "cannot import '{}' from '{}'"
            raise ImportError(msg.format(name, src_name))


def deprecate_linked(check=None, msg=None):
    """Deprecate an entire module if used through a link.

    This function must be called in the global context of the new module.

    :param check: Must be a module name to check, it must be part of the
           actual module name.  If not given 'xotl.tools.future' is assumed.

    For example::

      >>> from xotl.tools.deprecation import deprecate_linked
      >>> deprecate_linked()
      >>> del deprecate_linked

    """
    import inspect

    check = check or "xotl.tools.future"
    frame = inspect.currentframe().f_back
    try:
        name = frame.f_globals.get("__name__")
    finally:
        # As recommended in Python's documentation to avoid memory leaks
        del frame
    if check not in name:
        if msg is None:
            msg = (
                '"{}" module is now deprecated and it will be removed; use '
                'the one in "{}" instead.'
            ).format(name, check)
        warnings.warn(msg, stacklevel=2)


def deprecate_module(replacement, msg=None):
    """Deprecate an entire module.

    This function must be called in the global context of the deprecated
    module.

    :param replacement: The name of replacement module.

    For example::

      >>> from xotl.tools.deprecation import deprecate_module
      >>> deprecate_module('xotl.tools.symbols')
      >>> del deprecate_module

    """
    import inspect

    frame = inspect.currentframe().f_back
    try:
        name = frame.f_globals.get("__name__")
    finally:
        # As recommended in Python's documentation to avoid memory leaks
        del frame
    if msg is None:
        msg = (
            '"{}" module is now deprecated and it will be removed; ' 'use "{}" instead.'
        ).format(name, replacement)
    if msg:
        warnings.warn(msg, stacklevel=2)


@deprecated(import_deprecated)
def inject_deprecated(funcnames, source, target=None):
    """Injects a set of functions from a module into another.

    The functions will be marked as deprecated in the target module.

    :param funcnames: function names to take from the source module.

    :param source: the module where the functions resides.

    :param target: the module that will contains the deprecated functions.  If
           ``None`` will be the module calling this function.

    This function is provided for easing the deprecation of whole modules and
    should not be used to do otherwise.

    .. deprecated:: 1.8.0  Use `import_deprecated`:func:.

    """
    if not target:
        import sys

        frame = sys._getframe(1)
        try:
            target_locals = frame.f_locals
        finally:
            # As recommended to avoid memory leaks
            del frame
    else:
        # FIX: @manu, there is a consistency error here, 'target_locals' is
        # never assigned
        pass
    for targetname in funcnames:
        unset = object()
        target = getattr(source, targetname, unset)
        if target is not unset:
            testclasses = (types.FunctionType, types.LambdaType, type)
            if isinstance(target, testclasses):
                replacement = source.__name__ + "." + targetname
                module_name = target_locals.get("__name__", None)
                target_locals[targetname] = deprecated(
                    replacement, DEFAULT_MSG, module_name
                )(target)
            else:
                target_locals[targetname] = target
        else:
            warnings.warn(
                "{targetname} was expected to be in {source}".format(
                    targetname=targetname, source=source.__name__
                ),
                stacklevel=2,
            )


def deprecated_alias(f, **kwargs):
    """Declare a deprecated alias.

    This is roughly the same as ``deprecated(f)(f)``; which is makes it
    convenient to give a better name to an already released function `f`,
    while keeping the old name as a deprecated alias.

    .. versionadded:: 2.1.0

    """
    return deprecated(f, **kwargs)(f)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#
"""Functional tools for functions that returns iterators (generators, etc.)

.. warning:: This module is experimental.  It may be removed completely, moved
   or otherwise changed.

"""
from typing import Callable, Iterable, TypeVar
from functools import reduce

X = TypeVar("X")
Y = TypeVar("Y")
Z = TypeVar("Z")
T = TypeVar("T")


def kleisli_compose(*fs: Callable[[T], Iterable[T]]) -> Callable[[T], Iterable[T]]:
    """The Kleisli composition operator (right-to-left version).

    For two functions, ``kleisli_compose(g, f)`` returns::

       lambda x: (z for y in f(x) for z in g(y))

    In general this is, ``reduce(_compose, fs, lambda x: [x])``; where
    ``_compose`` is the lambda for two arguments.

    .. note:: Despite name (Kleisli), Python does not have a true Monad_
       type-class.  So this function works with functions taking a single
       argument and returning an iterator -- it also works with iterables.

    .. _Monad: https://en.wikipedia.org/wiki/Monad_(functional_programming)

    .. versionadded:: 1.9.6
    .. versionchanged:: 1.9.7 Name changed to ``kleisli_compose``.

    .. warning:: You may want to use `kleisli_compose_foldl`:func: which
       matches the order semantics of the functional kleisli composition
       ``>=>``.

    """

    def _kleisli_compose(
        g: Callable[[Y], Iterable[Z]], f: Callable[[X], Iterable[Y]]
    ) -> Callable[[X], Iterable[Z]]:
        # (>>.) :: Monad m => (b -> m c) -> (a -> m b) -> a -> m c
        # g >>. f = \x -> f x >>= g
        #
        # In the list monad:
        #
        # g >>. f = \x -> concat (map g (f x))
        return lambda x: (z for y in f(x) for z in g(y))

    if len(fs) == 2:
        # optimize a bit so that we can avoid the 'lambda x: [x]' for common
        # cases.
        return _kleisli_compose(*fs)
    else:
        return reduce(_kleisli_compose, fs, lambda x: iter([x]))


def kleisli_compose_foldl(
    *fs: Callable[[T], Iterable[T]]
) -> Callable[[T], Iterable[T]]:
    """Same as `kleisli_compose`:func: but composes left-to-right.

    Examples:

      >>> s15 = lambda s: tuple(s + str(i) for i in range(1, 5))
      >>> s68 = lambda s: tuple(s + str(i) for i in range(6, 8))

      >>> # kleisli_compose produces "6" >>= 1, 2, 3, 4; and then "7" >>= 1, 2, 3, 4
      >>> list(kleisli_compose(s15, s68)(""))
      ['61', '62', '63', '64', '71', '72', '73', '74']

      >>> list(kleisli_compose_foldl(s15, s68)(""))
      ['16', '17', '26', '27', '36', '37', '46', '47']

    If the operation is non-commutative (as the string concatenation) you end
    up with very different results.

      >>> n15 = lambda s: tuple(s + i for i in range(1, 5))
      >>> n68 = lambda s: tuple(s + i for i in range(6, 8))

      >>> list(kleisli_compose(n15, n68)(0))
      [7, 8, 9, 10, 8, 9, 10, 11]

      >>> list(kleisli_compose_foldl(n15, n68)(0))
      [7, 8, 8, 9, 9, 10, 10, 11]

    If the operation is commutative you get the same *set* of results, but the
    order may be different.

    """
    # This basically the same as _kleisli_compose above but with f and g
    # swapped.
    #
    # In our derivation of
    def _kleisli_compose_foldl(f, g):
        return lambda x: (z for y in f(x) for z in g(y))

    if len(fs) == 2:
        # optimize a bit so that we can avoid the 'lambda x: [x]' for common
        # cases.
        return _kleisli_compose_foldl(*fs)
    else:
        return reduce(_kleisli_compose_foldl, fs, lambda x: iter([x]))



# -*- encoding: utf-8 -*-

# Most of the code was taken from the Python 3.3 code base.
#
# Copyright (c) 2001-2012 Python Software Foundation.  All rights reserved.
#

import unittest
from xotl.tools.future.textwrap import wrap, fill, dedent, indent


class BaseTestCase(unittest.TestCase):
    """Parent class with utility methods for textwrap tests."""

    def show(self, textin):
        if isinstance(textin, list):
            result = []
            for i in range(len(textin)):
                result.append("  %d: %r" % (i, textin[i]))
            result = "\n".join(result)
        elif isinstance(textin, str):
            result = "  %s\n" % repr(textin)
        return result

    def check(self, result, expect):
        msg = "expected:\n%s\nbut got:\n%s" % (self.show(expect), self.show(result))
        self.assertEqual(result, expect, msg)

    def check_wrap(self, text, width, expect, **kwargs):
        result = wrap(text, width, **kwargs)
        self.check(result, expect)

    def check_split(self, text, expect):
        result = self.wrapper._split(text)
        self.assertEqual(
            result, expect, "\nexpected %r\n" "but got  %r" % (expect, result)
        )


class IndentTestCases(BaseTestCase):

    # called before each test method
    def setUp(self):
        self.text = """\
This paragraph will be filled, first without any indentation,
and then with some (including a hanging indent)."""

    def test_fill(self):
        # Test the fill() method

        expect = """\
This paragraph will be filled, first
without any indentation, and then with
some (including a hanging indent)."""

        result = fill(self.text, 40)
        self.check(result, expect)

    def test_initial_indent(self):
        # Test initial_indent parameter

        expect = [
            "     This paragraph will be filled,",
            "first without any indentation, and then",
            "with some (including a hanging indent).",
        ]
        result = wrap(self.text, 40, initial_indent="     ")
        self.check(result, expect)

        expect = "\n".join(expect)
        result = fill(self.text, 40, initial_indent="     ")
        self.check(result, expect)

    def test_subsequent_indent(self):
        # Test subsequent_indent parameter
        expect = """\
  * This paragraph will be filled, first
    without any indentation, and then
    with some (including a hanging
    indent)."""

        result = fill(self.text, 40, initial_indent="  * ", subsequent_indent="    ")
        self.check(result, expect)


# Despite the similar names, DedentTestCase is *not* the inverse
# of IndentTestCase!
class DedentTestCase(unittest.TestCase):
    def assertUnchanged(self, text):
        """assert that dedent() has no effect on 'text'"""
        self.assertEqual(text, dedent(text))

    def test_dedent_nomargin(self):
        # No lines indented.
        text = "Hello there.\nHow are you?\nOh good, I'm glad."
        self.assertUnchanged(text)

        # Similar, with a blank line.
        text = "Hello there.\n\nBoo!"
        self.assertUnchanged(text)

        # Some lines indented, but overall margin is still zero.
        text = "Hello there.\n  This is indented."
        self.assertUnchanged(text)

        # Again, add a blank line.
        text = "Hello there.\n\n  Boo!\n"
        self.assertUnchanged(text)

    def test_dedent_even(self):
        # All lines indented by two spaces.
        text = "  Hello there.\n  How are ya?\n  Oh good."
        expect = "Hello there.\nHow are ya?\nOh good."
        self.assertEqual(expect, dedent(text))

        # Same, with blank lines.
        text = "  Hello there.\n\n  How are ya?\n  Oh good.\n"
        expect = "Hello there.\n\nHow are ya?\nOh good.\n"
        self.assertEqual(expect, dedent(text))

        # Now indent one of the blank lines.
        text = "  Hello there.\n  \n  How are ya?\n  Oh good.\n"
        expect = "Hello there.\n\nHow are ya?\nOh good.\n"
        self.assertEqual(expect, dedent(text))

    def test_dedent_uneven(self):
        # Lines indented unevenly.
        text = """\
        def foo():
            while 1:
                return foo
        """
        expect = """\
def foo():
    while 1:
        return foo
"""
        self.assertEqual(expect, dedent(text))

        # Uneven indentation with a blank line.
        text = "  Foo\n    Bar\n\n   Baz\n"
        expect = "Foo\n  Bar\n\n Baz\n"
        self.assertEqual(expect, dedent(text))

        # Uneven indentation with a whitespace-only line.
        text = "  Foo\n    Bar\n \n   Baz\n"
        expect = "Foo\n  Bar\n\n Baz\n"
        self.assertEqual(expect, dedent(text))

    # dedent() should not mangle internal tabs
    def test_dedent_preserve_internal_tabs(self):
        text = "  hello\tthere\n  how are\tyou?"
        expect = "hello\tthere\nhow are\tyou?"
        self.assertEqual(expect, dedent(text))

        # make sure that it preserves tabs when it's not making any
        # changes at all
        self.assertEqual(expect, dedent(expect))

    # dedent() should not mangle tabs in the margin (i.e.
    # tabs and spaces both count as margin, but are *not*
    # considered equivalent)
    def test_dedent_preserve_margin_tabs(self):
        text = "  hello there\n\thow are you?"
        self.assertUnchanged(text)

        # same effect even if we have 8 spaces
        text = "        hello there\n\thow are you?"
        self.assertUnchanged(text)

        # dedent() only removes whitespace that can be uniformly removed!
        text = "\thello there\n\thow are you?"
        expect = "hello there\nhow are you?"
        self.assertEqual(expect, dedent(text))

        text = "  \thello there\n  \thow are you?"
        self.assertEqual(expect, dedent(text))

        text = "  \t  hello there\n  \t  how are you?"
        self.assertEqual(expect, dedent(text))

        text = "  \thello there\n  \t  how are you?"
        expect = "hello there\n  how are you?"
        self.assertEqual(expect, dedent(text))


# Test textwrap.indent
class IndentTestCase(unittest.TestCase):
    # The examples used for tests. If any of these change, the expected
    # results in the various test cases must also be updated.
    # The roundtrip cases are separate, because textwrap.dedent doesn't
    # handle Windows line endings
    ROUNDTRIP_CASES = (
        # Basic test case
        "Hi.\nThis is a test.\nTesting.",
        # Include a blank line
        "Hi.\nThis is a test.\n\nTesting.",
        # Include leading and trailing blank lines
        "\nHi.\nThis is a test.\nTesting.\n",
    )
    CASES = ROUNDTRIP_CASES + (
        # Use Windows line endings
        "Hi.\r\nThis is a test.\r\nTesting.\r\n",
        # Pathological case
        "\nHi.\r\nThis is a test.\n\r\nTesting.\r\n\n",
    )

    def test_indent_nomargin_default(self):
        # indent should do nothing if 'prefix' is empty.
        for text in self.CASES:
            self.assertEqual(indent(text, ""), text)

    def test_indent_nomargin_explicit_default(self):
        # The same as test_indent_nomargin, but explicitly requesting
        # the default behaviour by passing None as the predicate
        for text in self.CASES:
            self.assertEqual(indent(text, "", None), text)

    def test_indent_nomargin_all_lines(self):
        # The same as test_indent_nomargin, but using the optional
        # predicate argument
        predicate = lambda line: True
        for text in self.CASES:
            self.assertEqual(indent(text, "", predicate), text)

    def test_indent_no_lines(self):
        # Explicitly skip indenting any lines
        predicate = lambda line: False
        for text in self.CASES:
            self.assertEqual(indent(text, "    ", predicate), text)

    def test_roundtrip_spaces(self):
        # A whitespace prefix should roundtrip with dedent
        for text in self.ROUNDTRIP_CASES:
            self.assertEqual(dedent(indent(text, "    ")), text)

    def test_roundtrip_tabs(self):
        # A whitespace prefix should roundtrip with dedent
        for text in self.ROUNDTRIP_CASES:
            self.assertEqual(dedent(indent(text, "\t\t")), text)

    def test_roundtrip_mixed(self):
        # A whitespace prefix should roundtrip with dedent
        for text in self.ROUNDTRIP_CASES:
            self.assertEqual(dedent(indent(text, " \t  \t ")), text)

    def test_indent_default(self):
        # Test default indenting of lines that are not whitespace only
        prefix = "  "
        expected = (
            # Basic test case
            "  Hi.\n  This is a test.\n  Testing.",
            # Include a blank line
            "  Hi.\n  This is a test.\n\n  Testing.",
            # Include leading and trailing blank lines
            "\n  Hi.\n  This is a test.\n  Testing.\n",
            # Use Windows line endings
            "  Hi.\r\n  This is a test.\r\n  Testing.\r\n",
            # Pathological case
            "\n  Hi.\r\n  This is a test.\n\r\n  Testing.\r\n\n",
        )
        for text, expect in zip(self.CASES, expected):
            self.assertEqual(indent(text, prefix), expect)

    def test_indent_explicit_default(self):
        # Test default indenting of lines that are not whitespace only
        prefix = "  "
        expected = (
            # Basic test case
            "  Hi.\n  This is a test.\n  Testing.",
            # Include a blank line
            "  Hi.\n  This is a test.\n\n  Testing.",
            # Include leading and trailing blank lines
            "\n  Hi.\n  This is a test.\n  Testing.\n",
            # Use Windows line endings
            "  Hi.\r\n  This is a test.\r\n  Testing.\r\n",
            # Pathological case
            "\n  Hi.\r\n  This is a test.\n\r\n  Testing.\r\n\n",
        )
        for text, expect in zip(self.CASES, expected):
            self.assertEqual(indent(text, prefix, None), expect)

    def test_indent_all_lines(self):
        # Add 'prefix' to all lines, including whitespace-only ones.
        prefix = "  "
        expected = (
            # Basic test case
            "  Hi.\n  This is a test.\n  Testing.",
            # Include a blank line
            "  Hi.\n  This is a test.\n  \n  Testing.",
            # Include leading and trailing blank lines
            "  \n  Hi.\n  This is a test.\n  Testing.\n",
            # Use Windows line endings
            "  Hi.\r\n  This is a test.\r\n  Testing.\r\n",
            # Pathological case
            "  \n  Hi.\r\n  This is a test.\n  \r\n  Testing.\r\n  \n",
        )
        predicate = lambda line: True
        for text, expect in zip(self.CASES, expected):
            self.assertEqual(indent(text, prefix, predicate), expect)

    def test_indent_empty_lines(self):
        # Add 'prefix' solely to whitespace-only lines.
        prefix = "  "
        expected = (
            # Basic test case
            "Hi.\nThis is a test.\nTesting.",
            # Include a blank line
            "Hi.\nThis is a test.\n  \nTesting.",
            # Include leading and trailing blank lines
            "  \nHi.\nThis is a test.\nTesting.\n",
            # Use Windows line endings
            "Hi.\r\nThis is a test.\r\nTesting.\r\n",
            # Pathological case
            "  \nHi.\r\nThis is a test.\n  \r\nTesting.\r\n  \n",
        )
        predicate = lambda line: not line.strip()
        for text, expect in zip(self.CASES, expected):
            self.assertEqual(indent(text, prefix, predicate), expect)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

"""Extends the standard `types` module.

Standard module defines names for all type symbols known in the standard
interpreter.

This modules mirrors all the functions (and, in general, objects) from the
standard library module `types`:mod:; but it also includes several new types
and type-related functions.

In Jython and PyPy, `MemberDescriptorType` is identical to
`GetSetDescriptorType`; to mantain compatibility in some `xotl.tools` code,
they are differentiated in this module.

"""
from typing import TypeVar
from typing_extensions import Protocol

from types import *  # noqa
import types as _stdlib  # noqa

from xotl.tools.deprecation import deprecated

from xotl.tools.symbols import Unset as _unset
from collections.abc import Mapping


try:
    from types import __all__  # noqa

    __all__ = list(__all__)
except ImportError:
    # Python 3.3 don't implement '__all__' for 'string' module.
    __all__ = [name for name in dir(_stdlib) if not name.startswith("_")]

try:
    NoneType = _stdlib.NoneType  # noqa
except AttributeError:
    try:
        # In PyPy3 'NoneType' is a built-in
        from builtins import NoneType  # noqa
    except ImportError:
        NoneType = type(None)
    __all__.append("NoneType")

try:
    # It is maintained in this module for perhaps using it in `mypy`.
    EllipsisType  # noqa
except NameError:
    EllipsisType = type(Ellipsis)
    __all__.append("EllipsisType")

# Check Jython and PyPy peculiarity
if MemberDescriptorType is GetSetDescriptorType:  # noqa

    class _foo:
        __slots__ = "bar"

    MemberDescriptorType = type(_foo.bar)
    del _foo

FuncTypes = tuple(
    {
        FunctionType,
        MethodType,
        LambdaType,  # noqa
        BuiltinFunctionType,
        BuiltinMethodType,
    }
)  # noqa

func_types = FuncTypes  # Just an alias

from types import _calculate_meta  # noqa


import re

RegexPattern = type(re.compile(""))
del re


def _get_mro_attr(target, name, *default):
    """Get a named attribute from a type.

    Similar to `getattr` but looking in the MRO dictionaries for the type.
    Used internally in this module.

    For example::

      >>> class A(SimpleNamespace):
      ...     x = 12
      ...     y = 34

      >>> class B(A):
      ...     y = 56

      >>> b = B(x=1, y=2)

      >>> _get_mro_attr(b, 'x')
      12

      >>> _get_mro_attr(b, 'y')
      56

    """
    from xotl.tools.future.inspect import _static_getmro
    from xotl.tools.params import check_default, Undefined

    # force type
    target = target if isinstance(target, type) else type(target)
    target_mro = _static_getmro(target)
    cls = next((c for c in target_mro if name in c.__dict__), _unset)
    if cls is not _unset:
        return cls.__dict__[name]
    elif check_default()(*default) is not Undefined:
        return default[0]
    else:
        msg = "'{}' type has no attribute '{}'"
        raise AttributeError(msg.format(target, name))


@deprecated(_get_mro_attr)
class mro_dict(Mapping):
    """Utility behaving like a read-only dict of `target` MRO attributes.

    Used internally in this module.

    For example::

      >>> class A:
      ...     x = 12
      ...     y = 34

      >>> class B(A):
      ...     y = 56
      ...     z = 78

      >>> d = mro_dict(B)

      >>> d['x']
      12

      >>> d['y']
      56

      >>> d['z']
      78

    .. deprecated:: 1.8.4

    """

    __slots__ = ("_probes", "_keys")

    def __init__(self, target):
        from xotl.tools.future.inspect import _static_getmro

        type_ = target if isinstance(target, type) else type(target)
        target_mro = _static_getmro(type_)
        self._probes = tuple(c.__dict__ for c in target_mro)
        self._keys = set()

    def __getitem__(self, name):
        from xotl.tools.objects import get_first_of

        result = get_first_of(self._probes, name, default=_unset)
        if result is not _unset:
            return result
        else:
            raise KeyError(name)

    def __iter__(self):
        if not self._keys:
            self._settle_keys()
        return iter(self._keys)

    def __len__(self):
        if not self._keys:
            self._settle_keys()
        return len(self._keys)

    def _settle_keys(self):
        for probe in self._probes:
            for key in probe:
                if key not in self._keys:
                    self._keys.add(key)


@deprecated("None", '"mro_get_value_list" will be removed.')
def mro_get_value_list(cls, name):
    """Return a list with all `cls` class attributes in MRO.

    .. deprecated:: 1.8.4

    """
    return list(mro_get_full_mapping(cls, name).values())


@deprecated("None", '"mro_get_full_mapping" will be removed.')
def mro_get_full_mapping(cls, name):
    """Return a dictionary with all items from `cls` in MRO.

    All values corresponding to `name` must be valid mappings.

    .. deprecated:: 1.8.4

    """
    from xotl.tools.future.inspect import _static_getmro

    cls = cls if isinstance(cls, type) else type(cls)  # force type
    mro = _static_getmro(cls)
    return {t: t.__dict__[name] for t in mro if name in t.__dict__}


@deprecated("``iter(maybe)`` in an exception management block.")
def is_iterable(maybe):
    """Returns True if `maybe` is an iterable object.

    e.g. implements the `__iter__` method::

        >>> is_iterable('all strings are iterable')
        True

        # Numbers are not
        >>> is_iterable(1)
        False

        >>> is_iterable(range(1))
        True

        >>> is_iterable({})
        True

        >>> is_iterable(tuple())
        True

        >>> is_iterable(set())
        True

    .. deprecated:: 1.8.4

    """
    try:
        iter(maybe)
    except TypeError:
        return False
    else:
        return True


_is_collection_replacement = """::
    from xotl.tools.values.simple import collection, nil
    collection(avoid=Mapping)(maybe) is not nil
"""


@deprecated(_is_collection_replacement)
def is_collection(maybe):
    """Test `maybe` to see if it is a tuple, a list, a set or a generator
    function.

    It returns False for dictionaries and strings::

        >>> is_collection('all strings are iterable')
        False

        # Numbers are not
        >>> is_collection(1)
        False

        >>> is_collection(range(1))
        True

        >>> is_collection({})
        False

        >>> is_collection(tuple())
        True

        >>> is_collection(set())
        True

        >>> is_collection(a for a in range(100))
        True

    .. versionchanged:: 1.5.5 UserList are collections.

    .. deprecated:: 1.8.4

    """
    from xotl.tools.values.simple import logic_collection_coerce, nil

    return logic_collection_coerce(maybe) is not nil


@deprecated("``isinstance(maybe, Mapping)``")
def is_mapping(maybe):
    """Test `maybe` to see if it is a valid mapping.

    .. deprecated:: 1.8.4

    """
    return isinstance(maybe, Mapping)


@deprecated('``maybe + ""`` in an exception management block.')
def is_string_like(maybe):
    """Returns True if `maybe` acts like a string.

    .. deprecated:: 1.8.4

    """
    try:
        maybe + ""
    except TypeError:
        return False
    else:
        return True


@deprecated("None", '"is_scalar" will be removed.')
def is_scalar(maybe):
    """Returns if `maybe` is not not an iterable or a string.

    .. deprecated:: 1.8.4

    """
    from collections.abc import Iterable

    return isinstance(maybe, str) or not isinstance(maybe, Iterable)


def is_staticmethod(cls, name):
    """Returns true if a `method` is a static method.

    :param cls: The class or object that holds the method.

    :param name: The name of the method.

    When a static-method is declared, you can not test that condition using
    the traditional way::

      >>> class Foo:
      ...     @staticmethod
      ...     def bar():
      ...         pass

      >>> isinstance(Foo.bar, staticmethod)
      False

    Using this function::

      >>> is_staticmethod(Foo, 'bar')
      True

    """
    desc = _get_mro_attr(cls, name)
    return isinstance(desc, staticmethod)


def is_classmethod(cls, name):
    """Returns if a `method` is a class method.

    :param cls: The class or object that holds the method.

    :param name: The name of the method.

    When a class-method is declared, you can not test that condition using the
    traditional way::

      >>> class Foo:
      ...     @classmethod
      ...     def bar(cls):
      ...         pass

      >>> isinstance(Foo.bar, classmethod)
      False

    Using this function::

      >>> is_classmethod(Foo, 'bar')
      True

    """
    desc = _get_mro_attr(cls, name)
    return isinstance(desc, classmethod)


def is_instancemethod(cls, name):
    """Returns if a `method` is neither a static nor a class method.

    :param cls: The class or object that holds the method.

    :param name: The name of the method.

    To find out if a method is "normal", ``isinstance(obj.method,
    MethodType)`` can't be used::

      >>> class Foobar:
      ...     @classmethod
      ...     def cm(cls):
      ...         pass
      ...     def im(self):
      ...         pass

      >>> isinstance(Foobar.cm, MethodType)
      True

    Using this function::

      >>> is_instancemethod(Foobar, 'im')
      True

      >>> is_instancemethod(Foobar, 'cm')
      False

    """
    desc = _get_mro_attr(cls, name)
    return isinstance(desc, FunctionType)  # noqa


@deprecated("``isinstance(maybe, ModuleType)``")
def is_module(maybe):
    """Returns True if `maybe` is a module.

    .. deprecated:: 1.8.4

    """
    return isinstance(maybe, ModuleType)  # noqa


@deprecated("``all(isinstance(obj, types) for obj in subjects)``")
def are_instances(*args):
    """Return True if every `subject` is an instance of (any) `types`.

    :param subjects: All but last positional arguments.  Are the objects
        required to be instances of `types`.

    :param types: The last positional argument.  Either a single type or a
       sequence of types.  This must meet the conditions on the last
       argument of `isinstance`:func:.

    :returns: True or False.  True if for every `subject`,
       ``isinstance(subject, types)`` is True.  Otherwise, False.

    If no `subjects` are provided return True::

        >>> are_instances(int)
        True

    .. seealso:: The function `no_instances`:func: allows to test for
                 subjects not being instances of types.

    .. deprecated:: 1.8.4

    """
    from xotl.tools.params import check_count

    check_count(args, 1, caller="are_instances")
    *subjects, types = args
    if not subjects:
        isinstance(None, types)  # HACK: always validate `types`.
    return all(isinstance(subject, types) for subject in subjects)


@deprecated("``all(not isinstance(obj, types) for obj in subjects)``")
def no_instances(*args):
    """Return True if every `subject` is **not** an instance of (neither)
    `types`.

    :param subjects: All but last positional arguments.  Are the objects
           required not to be instances of `types`.

    :param types: The last positional argument.  Either a single type or a
           sequence of types.  This must meet the conditions on the last
           argument of `isinstance`:func:.

    :returns: True or False.  True if for every `subject`,
              ``isinstance(subject, types)`` is False.  Otherwise, False.

    If no `subjects` are provided return True::

        >>> no_instances(int)
        True

    .. note:: This is not the same as ``not are_instances(...)``.

       This function requires that *no* subject is an instance of `types`.
       Negating `are_instances`:func: would be True if *any* subject is
       not an instance of `types`.

    .. deprecated:: 1.8.4

    """
    from xotl.tools.params import check_count

    check_count(args, 1, caller="no_instances")
    *subjects, types = args
    if not subjects:
        isinstance(None, types)  # HACK: always validate `types`.
    return all(not isinstance(subject, types) for subject in subjects)


class TypeClass(Protocol):
    pass


# fmt: off
class EqTypeClass(TypeClass, Protocol):  # pragma: no cover
    def __eq__(self, other) -> bool: ...
    def __ne__(self, other) -> bool: ...


class OrdTypeClass(EqTypeClass, Protocol):  # pragma: no cover
    def __le__(self, other) -> bool: ...
    def __lt__(self, other) -> bool: ...
    def __ge__(self, other) -> bool: ...
    def __gt__(self, other) -> bool: ...
# fmt: on

TEq = TypeVar("TEq", bound=EqTypeClass)
TOrd = TypeVar("TOrd", bound=OrdTypeClass)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

import pytest


def test_regression_Command_repr():
    from xotl.tools.cli import Command

    class MyCommand(Command):
        def run(self):
            pass

    class Hidden(object):
        def run(self):
            pass

    Command.register(Hidden)

    cmd = MyCommand()
    assert repr(cmd) != ""
    registry = Command.registry
    assert registry.get("my-command")
    assert registry.get("hidden")


def test_can_actually_run_the_help():
    from xotl.tools.cli.app import main

    with pytest.raises(SystemExit):
        main(default="help")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#


import unittest
from xotl.tools.symbols import boolean


class BooleanTests(unittest.TestCase):
    def test_singletons(self):
        from sys import intern
        from xotl.tools.symbols import Unset

        foo = "Un"
        bar = "set"
        a = boolean("Unset")
        b = boolean(foo + "set")
        c = boolean("Un" + bar)

        # self.assertIs(intern(foo + bar), repr(Unset))
        # above started to fail in PyPy, changed to next.
        self.assertIs(intern(foo + bar), intern(repr(Unset)))
        self.assertIs(repr(a), repr(b))
        self.assertIs(a, Unset)
        self.assertIs(b, Unset)
        self.assertIs(c, Unset)

    def test_equality(self):
        a = boolean("false")
        b = boolean("true", True)
        self.assertEqual(a, False)
        self.assertEqual(b, True)
        self.assertEqual(not b, False)

    def test_parse(self):
        a = boolean("false")
        b = boolean("true", True)
        c = boolean.parse(repr(a))
        self.assertIs(boolean.parse("false"), a)
        self.assertIs(boolean.parse("true"), b)
        self.assertIs(a, c)

    def test_int_compatibility(self):
        a = boolean("false")
        b = boolean("true", True)
        self.assertEqual(a + 1, 1)
        self.assertEqual(b + 1, 2)

    def test_comments(self):
        a = boolean("false")
        value = "%s    # This is a comment" % a
        b = boolean.parse(value)
        self.assertIs(a, b)

    def test_symbols_are_pickable(self):
        import pickle
        from xotl.tools.symbols import Unset, Undefined

        for protocol in range(pickle.DEFAULT_PROTOCOL, pickle.HIGHEST_PROTOCOL + 1):
            self.assertIs(Unset, pickle.loads(pickle.dumps(Unset, protocol)))
            self.assertIs(Undefined, pickle.loads(pickle.dumps(Undefined, protocol)))


def test_symbols_is_importable():
    import sys

    modules = {
        mod: sys.modules[mod] for mod in sys.modules if mod.startswith("xotl.tools.")
    }
    for mod in modules:
        sys.modules.pop(mod)
    try:
        import xotl.tools.symbols  # noqa
    finally:
        for modname, mod in modules.items():
            sys.modules[modname] = mod



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#
from hypothesis import strategies as s, given, example


def test_fp_compose():
    from xotl.tools.fp.tools import identity, compose, pos_args, kw_args, full_args

    x, obj = 15, object()
    f, g, h = x.__add__, x.__mul__, x.__xor__

    def join(*args):
        if args:
            return " -- ".join(str(arg) for arg in args)
        # functions return 'None' when no explicit 'return' is issued.

    def plus2(value):
        return value + 2

    def plus2d(value):
        return {"stop": value + 2}

    def myrange(stop):
        return list(range(stop))

    assert compose(join, pos_args, myrange, plus2)(0) == "0 -- 1"
    assert compose(join, myrange, plus2)(0) == "[0, 1]"

    assert compose(join, myrange, kw_args, plus2d)(0) == "[0, 1]"
    assert compose(join, myrange, full_args.parse, plus2d)(0) == "[0, 1]"

    assert compose() is identity
    assert compose()(x) is x
    assert compose()(obj) is obj
    assert compose(f) is f
    assert compose(g, f)(x) == g(f(x))
    assert compose(h, g, f)(x) == h(g(f(x)))


def test_fp_compose_wrapable():
    from functools import wraps
    from xotl.tools.fp.tools import compose

    def wrapper():
        "X"
        pass

    res = wraps(wrapper)(compose(wrapper, lambda: None))
    assert res.__name__ == wrapper.__name__
    assert res.__doc__ == wrapper.__doc__
    assert res.__module__ == wrapper.__module__


def test_fp_tools():
    from xotl.tools.fp.tools import identity, compose

    x, obj = 15, object()
    f, g, h = x.__add__, x.__mul__, x.__xor__

    def join(*args):
        if args:
            return " -- ".join(str(arg) for arg in args)
        # functions return 'None' when no explicit 'return' is issued.

    def plus2(value):
        return value + 2

    def plus2d(value):
        return {"stop": value + 2}

    def myrange(stop):
        return list(range(stop))

    assert compose(join, myrange, plus2)(0) == "[0, 1]"
    assert compose() is identity
    assert compose()(x) is x
    assert compose()(obj) is obj
    assert compose(f) is f
    assert compose(g, f)(x) == g(f(x))
    assert compose(h, g, f)(x) == h(g(f(x)))

    c = compose(*((lambda y: lambda x: x + y)(i) for i in range(6)))
    for i in range(7):
        assert c[:i](0) == sum(range(i))


@given(s.integers(min_value=0, max_value=20))
@example(4)
def test_fp_kleisli_compose(n):
    from xotl.tools.fp.iterators import kleisli_compose

    def fullrange(n):
        "[0..n]"
        return range(n + 1)

    def odds(n):
        return [x for x in fullrange(n) if x % 2 != 0]

    odd_seqs = kleisli_compose(odds, fullrange)
    assert list(odd_seqs(n)) == [z for y in fullrange(n) for z in odds(y)]
    id_ = lambda x: [x]
    pad = (id_,) * n
    args = pad + (odds,) + pad + (fullrange,) + pad
    odd_seqs = kleisli_compose(*args)
    assert list(odd_seqs(n)) == [z for y in fullrange(n) for z in odds(y)]

    odd_seqs = kleisli_compose(fullrange, odds)
    assert list(odd_seqs(n)) == [z for y in odds(n) for z in fullrange(y)]
    args = pad + (fullrange,) + pad + (odds,) + pad
    odd_seqs = kleisli_compose(*args)
    assert list(odd_seqs(n)) == [z for y in odds(n) for z in fullrange(y)]


def test_fp_kleisli_compose4():
    from xotl.tools.fp.iterators import kleisli_compose

    def fullrange(n):
        "[0..n]"
        return range(n + 1)

    def odds(n):
        return [x for x in fullrange(n) if x % 2 != 0]

    id_ = lambda x: [x]
    odd_seqs = kleisli_compose(odds, fullrange)
    assert list(odd_seqs(4)) == [1, 1, 1, 3, 1, 3]
    odd_seqs = kleisli_compose(id_, odds, id_, id_, fullrange, id_)
    assert list(odd_seqs(4)) == [1, 1, 1, 3, 1, 3]

    odd_seqs = kleisli_compose(fullrange, odds)
    assert list(odd_seqs(4)) == [0, 1, 0, 1, 2, 3]
    odd_seqs = kleisli_compose(id_, fullrange, id_, id_, odds, id_)
    assert list(odd_seqs(4)) == [0, 1, 0, 1, 2, 3]


anything = (
    s.integers()
    | s.dictionaries(s.text(), s.integers())
    | s.text()
    | s.tuples(s.integers())
)
anyargs = s.tuples(anything)


@given(anything, anyargs)
def test_constant(val, args):
    from xotl.tools.fp.tools import constant

    fn = constant(val)
    assert fn(*args) is val


if __name__ == "__main__":
    # Allow to run this test as an independent python script, for example:
    #   monkeytype run tests/test_fp.py
    test_fp_compose()
    test_fp_compose_wrapable()
    test_fp_tools()
    test_fp_kleisli_compose()
    test_fp_kleisli_compose4()
    test_constant()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

"""Text wrapping and filling."""

from textwrap import *  # noqa
import textwrap as _stdlib
from textwrap import __all__  # noqa

__all__ = list(__all__)


def dedent(text, skip_firstline=False):
    r"""Remove any common leading whitespace from every line in text.

    This can be used to make triple-quoted strings line up with the left edge
    of the display, while still presenting them in the source code in indented
    form.

    Note that tabs and spaces are both treated as whitespace, but they are not
    equal: the lines ``"    hello"`` and ``"\thello"`` are considered to have
    no common leading whitespace.

    If `skip_firstline` is True, the first line is separated from the rest of
    the body.  This helps with docstrings that follow `257`:pep:.

    .. warning:: The `skip_firstline` argument is missing in standard library.

    """
    if skip_firstline:
        parts = text.split("\n", 1)
        if len(parts) > 1:
            subject, body = parts
        else:
            subject, body = parts[0], ""
        result = _stdlib.dedent(subject)
        if body:
            result += "\n" + _stdlib.dedent(body)
    else:
        result = _stdlib.dedent(text)
    return result



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

"""Advanced functional programming in Python.

.. note:: This module is in **EXPERIMENTAL** state, we encourage not to use it
   before declared stable.

"""



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

from hypothesis import strategies


@strategies.composite
def timespans(draw, dates=None, unbounds="any", always_valid=True):
    """A strategy that generates `xotl.tools.future.datetime.TimeSpan`:class:.

    This is a `hypothesis`_ strategy.

    `dates` should be None or generator of `datetime.date`:class: objects.
    It defaults to `hypothesis.strategies.dates`:func:

    `unbounds` should be one of the strings 'none', 'any', 'past', 'future'.
    If 'any' then the generated time span can be `unbound
    <xotl.tools.future.datetime.TimeSpan.unbound>`:attr:.  If 'past' it can be
    unbound to the past.  If 'future', it can be unbound to the future.  If
    'none', the generated time span will always be bound.  In all cases that
    generate unbound time spans, we can also generate bound time spans.

    If `always_valid` is True all generated time spans will be `valid
    <xotl.tools.future.datetime.TimeSpan.valid>`:attr:.  Otherwise we may
    generate invalid ones.

    Usage::

       >>> from hypothesis import given
       >>> from xotl.tools.testing.datetime import timespans

       >>> @given(timespans())
       ... def test_timespan(ts):
       ...    pass

    .. _hypothesis: https://hypothesis.readthedocs.io/

    .. versionadded:: 1.8.2

    """
    from xotl.tools.future.datetime import TimeSpan

    if dates is None:
        dates = strategies.dates()
    maybe = strategies.none() | dates
    if unbounds in ("any", "past", "future", "none"):
        if unbounds in ("any", "past"):
            date1 = draw(maybe)
        else:
            date1 = draw(dates)
        if unbounds in ("any", "future"):
            date2 = draw(maybe)
        else:
            date2 = draw(dates)
    else:
        raise ValueError("unbounds should be one of 'any', 'past', or 'future'.")
    if date1 and date2 and always_valid:
        start1 = min(date1, date2)
        end1 = max(date1, date2)
    else:
        start1 = date1
        end1 = date2
    return TimeSpan(start_date=start1, end_date=end1)


@strategies.composite
def datetimespans(draw, dates=None, unbounds="any", always_valid=True):
    """A strategy that generates `xotl.tools.future.datetime.DateTimeSpan`:class:.

    This is a `hypothesis`_ strategy.

    `dates` should be None or generator of `datetime.datetime`:class: objects.
    It defaults to `hypothesis.strategies.datetimes`:func:

    `unbounds` should be one of the strings 'none', 'any', 'past', 'future'.
    If 'any' then the generated time span can be `unbound
    <xotl.tools.future.datetime.TimeSpan.unbound>`:attr:.  If 'past' it can be
    unbound to the past.  If 'future', it can be unbound to the future.  If
    'none', the generated time span will always be bound.  In all cases that
    generate unbound time spans, we can also generate bound time spans.

    If `always_valid` is True all generated time spans will be `valid
    <xotl.tools.future.datetime.TimeSpan.valid>`:attr:.  Otherwise we may
    generate invalid ones.

    Usage::

       >>> from hypothesis import given
       >>> from xotl.tools.testing.datetime import datetimespans

       >>> @given(datetimespans())
       ... def test_datetimespan(dts):
       ...    pass

    .. _hypothesis: https://hypothesis.readthedocs.io/

    .. versionadded:: 1.9.7

    """
    from xotl.tools.future.datetime import DateTimeSpan

    if dates is None:
        dates = strategies.datetimes()
    maybe = strategies.none() | dates
    if unbounds in ("any", "past", "future", "none"):
        if unbounds in ("any", "past"):
            date1 = draw(maybe)
        else:
            date1 = draw(dates)
        if unbounds in ("any", "future"):
            date2 = draw(maybe)
        else:
            date2 = draw(dates)
    else:
        raise ValueError("unbounds should be one of 'any', 'past', or 'future'.")
    if date1 and date2 and always_valid:
        start1 = min(date1, date2)
        end1 = max(date1, date2)
    else:
        start1 = date1
        end1 = date2
    return DateTimeSpan(start_datetime=start1, end_datetime=end1)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#
import pytest
from xotl.tools.versions import python_version
from xotl.tools.future.collections import OrderedSmartDict

PYPY = python_version.pypy


def test_nameof():
    from xotl.tools.names import nameof, simple_name
    from collections import OrderedDict as sorted_dict

    assert nameof(sorted_dict) == "sorted_dict"
    assert nameof(sorted_dict, inner=True) == "OrderedDict"
    sd = sorted_dict(x=1, y=2)
    assert nameof(sd) == "sd"
    assert nameof(sd, typed=True) == "sorted_dict"
    assert nameof(sd, inner=True, typed=True) == "OrderedDict"
    s = "foobar"
    assert nameof(s, inner=True) == "foobar"
    # The following needs to be tested outside the assert, cause in Py3.3,
    # py.test rewrites the assert sentences and the local scope `nameof`
    # searched is not reached properly.
    passed = nameof("foobar") == "s"
    assert passed

    i = 1
    assert nameof(i) == "i"
    assert nameof(i, inner=True) == "1"
    assert nameof(i, typed=True) == "int"
    assert hex(id(sd)) in nameof(sd, inner=True)
    values = (None, True, False, BaseException, int, dict, object)
    names = [simple_name(v) for v in values]
    names.sort()
    assert names == ["BaseException", "False", "None", "True", "dict", "int", "object"]


def test_nameof_methods():
    from xotl.tools.names import nameof, simple_name

    class Foobar:
        def __init__(self):
            self.attr = "foobar"

        def first(self):
            pass

        @staticmethod
        def second():
            pass

        @classmethod
        def third(cls):
            pass

    obj = Foobar()
    attrs = (getattr(obj, n) for n in dir(obj) if not n.startswith("_"))
    attrs = (v for v in attrs if callable(v))
    names = nameof(*attrs)
    names.sort()
    assert names == ["first", "second", "third"]
    attrs = (getattr(obj, n) for n in dir(obj) if not n.startswith("_"))
    attrs = (v for v in attrs if callable(v))
    names = [simple_name(v, join=False) for v in attrs]
    names.sort()
    assert names == ["first", "second", "third"]


@pytest.mark.skipif(PYPY, reason="'OrderedDict' is in '_pypy_collections'")
def test_fullnameof():
    from xotl.tools.names import nameof, simple_name

    _name = "collections.OrderedDict"
    from collections import OrderedDict as sorted_dict

    assert nameof(sorted_dict, full=True) == "test_fullnameof.sorted_dict"
    assert nameof(sorted_dict, inner=True, full=True) == _name
    sd = sorted_dict(x=1, y=2)
    assert nameof(sd, full=True) == "test_fullnameof.sd"
    assert nameof(sd, typed=True, full=True) == "test_fullnameof.sorted_dict"
    assert nameof(sd, inner=True, typed=True, full=True) == _name
    assert simple_name(simple_name) == "xotl.tools.names.simple_name"
    assert simple_name(sd) == "collections.OrderedDict"


@pytest.mark.skipif(PYPY, reason="'OrderedDict' is in '_pypy_collections'")
def test_fullnameof_no_rename():
    from xotl.tools.names import nameof
    from collections import OrderedDict

    _full_name = "test_fullnameof_no_rename.OrderedDict"
    _name = "collections.OrderedDict"
    assert nameof(OrderedDict, full=True) == _full_name
    assert nameof(OrderedDict, inner=True, full=True) == _name


def test_module_level_name():
    from xotl.tools.names import nameof

    assert nameof(OrderedSmartDict) == "OrderedSmartDict"
    assert nameof(OrderedSmartDict, typed=True) == "OrderedSmartDict"


def test_module_level_name_isolated():
    from xotl.tools.names import nameof, simple_name

    full_name_1 = "tests.test_names.OrderedSmartDict"
    full_name_2 = "xotl.tools.future.collections.OrderedSmartDict"
    assert nameof(OrderedSmartDict, full=True) == full_name_1
    assert simple_name(OrderedSmartDict) == full_name_2



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

r"""Utilities for command-line interface (CLI) applications.

- `program_name`:func:\ : calculate the program name from "sys.argv[0]".

- `command_name`:func:\ : calculate command names using class names in lower
   case inserting a hyphen before each new capital letter.

"""


def hyphen_name(name, join_numbers=True):
    """Convert a name to a hyphened slug.

    Expects a `name` in Camel-Case.  All invalid characters (those invalid in
    Python identifiers) are ignored.  Numbers are joined with preceding part
    when `join_numbers` is True.

    For example::

      >>> hyphen_name('BaseNode') == 'base-node'
      True

      >> hyphen_name('--__ICQNámeP12_34Abc--') == 'icq-name-p12-34-abc'
      True

      >> hyphen_name('ICQNámeP12', join_numbers=False) == 'icq-name-p-12'
      True

    """
    import re
    from xotl.tools.string import force_ascii

    name = force_ascii(name)
    regex = re.compile("[^A-Za-z0-9]+")
    name = regex.sub("-", name)
    regex = re.compile("([A-Z]+|[a-z]+|[0-9]+|-)")
    all = regex.findall(name)
    i, count, parts = 0, len(all), []
    while i < count:
        part = all[i]
        if part != "-":
            upper = "A" <= part <= "Z"
            if upper:
                part = part.lower()
            j = i + 1
            if j < count and upper and "a" <= all[j] <= "z":
                aux = part[:-1]
                if aux:
                    parts.append(aux)
                part = part[-1] + all[j]
                i = j
                j += 1
            if j < count and "0" <= all[j] <= "9" and join_numbers:
                part = part + all[j]
                i = j
            parts.append(part)
        i += 1
    return "-".join(parts)


def program_name():
    """Calculate the program name from "sys.argv[0]"."""
    # TODO: Use 'argparse' standard (parser.prog)
    import sys
    from os.path import basename

    return basename(sys.argv[0])


def command_name(cls):
    """Calculate a command name from given class.

    Names are calculated putting class names in lower case and inserting
    hyphens before each new capital letter.  For example "MyCommand" will
    generate "my-command".

    It's defined as an external function because a class method don't apply to
    minimal commands (those with only the "run" method).

    Example::

        >>> class SomeCommand:
        ...     pass

        >>> command_name(SomeCommand) == 'some-command'
        True

    If the command class has an attribute `command_cli_name`, this will be
    used instead::

        >>> class SomeCommand:
        ...    command_cli_name = 'adduser'

        >>> command_name(SomeCommand) == 'adduser'
        True

    It's an error to have a non-string `command_cli_name` attribute::

        >>> class SomeCommand:
        ...    command_cli_name = None

        >>> command_name(SomeCommand)  # doctest: +ELLIPSIS
        Traceback (most recent call last):
           ...
        TypeError: Attribute 'command_cli_name' must be a string.

    """
    unset = object()
    names = ("command_cli_name", "__command_name__")
    i, res = 0, unset
    while i < len(names) and res is unset:
        name = names[i]
        res = getattr(cls, names[i], unset)
        if res is unset:
            i += 1
        elif not isinstance(res, str):
            raise TypeError("Attribute '{}' must be a string.".format(name))
    if res is unset:
        res = hyphen_name(cls.__name__)
    return res



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

"""Some generic value validators and regular expressions and validation
functions for several identifiers.

"""


# TODO: Check next import, it looks like one of the modules must be deprecated
from xotl.tools.validators.identifiers import (  # noqa
    is_valid_identifier,
    check_identifier,
    is_valid_full_identifier,
    is_valid_public_identifier,
    is_valid_slug,
)


def _adorn_checker_name(name):
    """Make more attractive or legible a checker name."""
    res = name.replace("_AND_", " & ")
    res = res.replace("_OR_", " | ")
    return res.replace("<lambda>", "<λ>")


def _get_checker_name(checker):
    """Return a nice name for a `checker`.

    A `checker` could be a type, a tuple of types, a callable or a list of
    other checkers.

    """
    l = lambda o: str("(%s)" % o.join(_get_checker_name(c) for c in checker))
    if isinstance(checker, list):
        return l("_AND_")
    elif isinstance(checker, tuple):
        return l("_OR_")
    else:
        from xotl.tools.future.inspect import safe_name

        res = safe_name(checker, affirm=True)
        if not isinstance(checker, type):
            assert callable(checker)
            if "lambda" in res:
                from inspect import getargspec

                args = getargspec(checker).args
                assert len(args) == 1
                res = str("%s(%s)" % (res, args[0]))
        return res


def is_type(cls):
    """Return a validator with the same name as the type given as argument
    `value`.

    :param cls: Class or type or tuple of several types.

    """

    def inner(obj):
        """Check is a value object is a valid instance of (%s)."""
        return isinstance(obj, cls)

    name = _get_checker_name(cls)
    inner.__name__ = name
    inner.__doc__ = inner.__doc__ % name
    return inner


# TODO: With this new function, `is_type` could be deprecated
# TODO: Migrate to a class
def predicate(*checkers, **kwargs):
    """Return a validation checker for types and simple conditions.

    :param checkers: A variable number of checkers; each one could be:

        - A type, or tuple of types, to test valid values with
          ``isinstance(value, checker)``

        - A set or mapping of valid values, the value is valid if contained in
          the checker.

        - A tuple of other inner checkers, if any of the checkers validates a
          value, the value is valid (OR).

        - A list of other inner checkers, all checkers must validate the value
          (AND).

        - A callable that receives the value and returns True if the value is
          valid.

        - ``True`` and ``False`` could be used as checkers always validating
          or invalidating the value.

        An empty list or no checker is synonym of ``True``, an empty tuple,
        set or mapping is synonym of ``False``.

    :param name: Keyword argument to be used in case of error; will be the
        argument of `ValueError` exception; could contain the placeholders
        ``{value}`` and ``{type}``; a default value is used if this argument
        is not given.

    :param force_name: Keyword argument to force a name if not given.

    In order to obtain good documentations, use proper names for functions and
    lambda arguments.

    With this function could be built real type checkers, for example::

      >>> is_valid_age = predicate((int, float), lambda age: 0 < age <= 120)
      >>> is_valid_age(100)
      True

      >>> is_valid_age(130)
      False

      >>> always_true = predicate(True)
      >>> always_true(False)
      True

      >>> always_false = predicate(False)
      >>> always_false(True)
      False

      >>> always_true = predicate()
      >>> always_true(1)
      True

      >>> always_true('any string')
      True

      >>> always_false = predicate(())
      >>> always_false(1)
      False

      >>> always_false('any string')
      False

    """
    from xotl.tools.symbols import boolean
    from xotl.tools.future.collections import Set, Mapping

    def inner(obj):
        """Check is `obj` is a valid instance for a set of checkers."""

        def valid(chk):
            if isinstance(chk, boolean):
                res = bool(chk)
            elif isinstance(chk, type):
                res = isinstance(obj, chk)
            elif isinstance(chk, tuple):
                if all(isinstance(c, type) for c in chk):
                    res = isinstance(obj, chk)
                else:
                    res = any(valid(c) for c in chk)
            elif isinstance(chk, list):
                res = all(valid(c) for c in chk)
            elif isinstance(chk, (Set, Mapping)):
                res = obj in chk
            else:
                res = chk(obj)
            return res

        # XXX: WTF, must be ``all(valid(chk) for chk in checkers)``
        return next((chk for chk in checkers if not valid(chk)), None) is None

    name = kwargs.get("name")
    if name is None and kwargs.get("force_name"):
        name = _get_checker_name(list(checkers))
    if name is not None:
        inner.__name__ = name
    return inner


def check(value, validator, msg=None):
    """Check a `value` with a `validator`.

    Argument `validator` could be a callable, a type, or a tuple of types.

    Return True if the value is valid.

    Examples::

      >>> check(1, int)
      True

      >>> check(10, lambda x: x <= 100, 'must be less than or equal to 100')
      True

      >>> check(11/2, (int, float))
      True

    """
    if isinstance(validator, (type, tuple)):
        checker = is_type(validator)
    else:
        checker = validator
    if checker(value):
        return True
    else:
        from xotl.tools.future.inspect import safe_name

        if not msg:
            # TODO: Use the name of validator with `inspect.getattr_static`
            # when `xotl.tools.future` is ready
            msg = 'Invalid value "%s" of type "%s"'
        msg = msg.format(value=value, type=safe_name(value, affirm=True))
        raise ValueError(msg)


# TODO: deprecate `check` in favor of `ok`.
def ok(value, *checkers, **kwargs):
    """Validate a value with several checkers.

    Return the value if it is Ok, or raises an `ValueError` exception if not.

    Arguments:

    :param value: the value to validate

    :param checkers: a variable number of checkers (at least one), each one
        could be a type, a tuple of types of a callable that receives the
        value and returns if the value is valid or not.  In order the value is
        considered valid, all checkers must validate the value.

    :param message: keyword argument to be used in case of error; will be the
           argument of `ValueError` exception; could contain the placeholders
           ``{value}`` and ``{type}``; a default value is used if this
           argument is not given.

    :param msg: an alias for "message"

    :param extra_checkers: In order to create validators using `partial`.
           Must be a tuple.

    Keyword arguments are not validated to be correct.

    This function could be used with type-definitions for arguments, see
    `xotl.tools.fp.prove.semantic.TypeCheck`:class:.

    Examples::

      >>> ok(1, int)
      1

      >>> ok(10, int, lambda x: x < 100, message='Must be integer under 100')
      10

      >>> ok(11/2, (int, float))
      5.5

      >>> ok(11/2, int, float)
      5.5

      >>> try:
      ...     res = ok(11/2, int)
      ... except ValueError:
      ...     res = '---'
      >>> res
      '---'

    """
    extra_checkers = kwargs.get("extra_checkers", ())
    pred = predicate(*(checkers + extra_checkers))
    if pred(value):
        return value
    else:
        from xotl.tools.future.itertools import multi_get as get
        from xotl.tools.future.inspect import safe_name

        msg = next(get(kwargs, "message", "msg"), "Invalid {type}: {value}!")
        msg = msg.format(value=value, type=safe_name(value, affirm=True))
        raise ValueError(msg)


def check_no_extra_kwargs(kwargs):
    """Check that no extra keyword arguments are still not processed.

    For example::

      >>> from xotl.tools.validators import check_no_extra_kwargs
      >>> def only_safe_arg(**kwargs):
      ...     safe = kwargs.pop('safe', False)
      ...     check_no_extra_kwargs(kwargs)
      ...     print('OK for safe:', safe)

    """
    if kwargs:
        plural = "" if len(kwargs) == 1 else "s"
        msg = 'Unexpected keyword argument%s: "%s"!'
        raise TypeError(msg % (plural, ", ".join(kwargs)))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

"""Prove validity of values - Base predicate classes.

A `predicate`:class: could combine two concepts (*validation* and
*conversion*) in a single callable object.

"""


class predicate:
    """Base class for value proves using logic predicates.

    A predicate could combine two operations on values: *validation*, and
    *conversion*.

    To signal that value as invalid, a predicate must return the special value
    `_wrong`.  Types work as in `isinstance`:func: standard function; callable
    functions mould a parameter value into a definitive form.

    To use normal functions as a callable predicate, use `SafeCheck`:class: or
    `LogicalCheck`:class` to wrap them.

    When using a list to combine explicitly the two concepts, result of the
    check part is considered Boolean (True or False), and the second part
    alwasy return a moulded value.

    When use a predicate, several definitions will be tried until one succeed.

    """

    __slots__ = ("inner",)

    def __new__(cls, *args):
        if cls is predicate:  # Parse the right sub-type
            count = len(args)
            if count == 0:
                msg = "{}() takes at least 1 argument (0 given)"
                raise TypeError(msg.format(cls.__name__))
            elif count == 1:
                arg = args[0]
                if isinstance(arg, cls):
                    return arg
                elif isinstance(arg, (tuple, type)):
                    return TypeCheck(arg)
                elif isinstance(arg, list):
                    return CheckAndCast(*arg)
                elif callable(arg):
                    return LogicalCheck(arg)
                else:
                    msg = "{}() can't parse a definition of type: {}"
                    aname = type(arg).__name__
                    raise TypeError(msg.format(cls.__name__, aname))
            else:
                return MultiCheck(*args)
        else:
            return super().__new__(cls)

    def __init__(self, *args):
        pass

    def __repr__(self):
        return str(self)


class TypeCheck(predicate):
    """Check if value is instance of given types."""

    __slots__ = ()

    def __new__(cls, *args):
        from xotl.tools.params import check_count

        check_count(len(args) + 1, 2, caller=cls.__name__)
        if len(args) == 1 and isinstance(args[0], tuple):
            args = args[0]
        if all(isinstance(arg, type) for arg in args):
            self = super().__new__(cls)
            self.inner = args
            return self
        else:
            wrong = (arg for arg in args if not isinstance(arg, type))
            wnames = ", or ".join(type(w).__name__ for w in wrong)
            msg = "`TypeCheck` allows only valid types, not: ({})"
            raise TypeError(msg.format(wnames))

    def __call__(self, value):
        from xotl.tools.fp.option import Just, Wrong

        ok = isinstance(value, self.inner)
        return (value if value else Just(value)) if ok else Wrong(value)

    def __str__(self):
        return self._str()

    def __crop__(self, max_width=None, canonical=False):
        """Calculate both string versions (small and normal)."""
        from xotl.tools.symbols import Undefined
        from xotl.tools.clipping import ELLIPSIS, DEFAULT_MAX_WIDTH

        if max_width is None:
            max_width = DEFAULT_MAX_WIDTH  # a big number for this
        start, end = "{}(".format(type(self).__name__), ")"
        borders_len = len(start) + len(end)
        sep = ", "
        res = ""
        items = iter(self.inner)
        ok = True
        while ok:
            item = next(items, Undefined)
            if item is not Undefined:
                if res:
                    res += sep
                aux = item.__name__
                if len(res) + len(aux) + borders_len <= max_width:
                    res += aux
                else:
                    res += ELLIPSIS
                    ok = False
            else:
                ok = False
        return "{}{}{}".format(start, res, end)


class NoneOrTypeCheck(TypeCheck):
    """Check if value is None or instance of given types."""

    __slots__ = ()

    def __call__(self, value):
        from xotl.tools.fp.option import Wrong

        if value is None:
            _types = self.inner
            i, res = 0, None
            while res is None and i < len(_types):
                try:
                    res = _types[i]()
                except Exception:
                    pass
                i += 1
            return res if res is not None else Wrong(None)
        else:
            return super().__call__(value)

    def __str__(self):
        aux = super().__str__()
        return "none-or-{}".format(aux)


class TypeCast(TypeCheck):
    """Cast a value to a correct type."""

    __slots__ = ()

    def __call__(self, value):
        from xotl.tools.fp.option import Just

        res = super().__call__(value)
        if not res:
            _types = self.inner
            i = 0
            while not res and i < len(_types):
                try:
                    res = _types[i](value)
                    if not res:
                        res = Just(res)
                except Exception:
                    pass
                i += 1
        return res

    def __str__(self):
        # FIX: change this
        aux = super(NoneOrTypeCheck, self).__str__()
        return "none-or-{}".format(aux)


class CheckAndCast(predicate):
    """Check if value, if valid cast it.

    Result value must be valid also.
    """

    __slots__ = ()

    def __new__(cls, check, cast):
        check = predicate(check)
        if callable(cast):
            self = super().__new__(cls)
            self.inner = (check, SafeCheck(cast))
            return self
        else:
            msg = '{}() expects a callable for cast, "{}" given'
            sname = type(self).__name__
            raise TypeError(msg.format(sname, type(cast).__name__))

    def __call__(self, value):
        from xotl.tools.fp.option import Wrong

        check, cast = self.inner
        aux = check(value)
        if aux:
            res = cast(value)
            if check(res):
                return res
        else:
            res = aux
        if isinstance(res, Wrong):
            return res
        else:
            return Wrong(value)

    def __str__(self):
        from xotl.tools.clipping import crop

        check, cast = self.inner
        fmt = "({}(…) if {}(…) else _wrong)"
        return fmt.format(crop(cast), check)


class FunctionalCheck(predicate):
    """Check if value is valid with a callable function."""

    __slots__ = ()

    def __new__(cls, check):
        # TODO: Change next, don't use isinstance
        if isinstance(check, predicate):
            return check
        elif callable(check):
            self = super().__new__(cls)
            self.inner = check
            return self
        else:
            msg = 'a functional check expects a callable but "{}" is given'
            raise TypeError(msg.format(type(check).__name__))

    def __str__(self):
        from xotl.tools.clipping import crop

        suffix = "check"
        kind = type(self).__name__.lower()
        if kind.endswith(suffix):
            kind = kind[: -len(suffix)]
        inner = crop(self.inner)
        return "{}({})()".format(kind, inner)


class LogicalCheck(FunctionalCheck):
    """Check if value is valid with a callable function."""

    __slots__ = ()

    def __call__(self, value):
        from xotl.tools.fp.option import Just, Wrong

        try:
            res = self.inner(value)
            if res:
                if isinstance(res, Just):
                    return res
                elif res is True:
                    return Just(value)
                else:
                    return res
            elif isinstance(res, Wrong):
                return res
            elif res is False or res is None:  # XXX: None?
                return Wrong(value)
            else:
                return Wrong(res)
        except Exception as error:
            return Wrong(error)


class SafeCheck(FunctionalCheck):
    """Return a wrong value only if function produce an exception."""

    __slots__ = ()

    def __call__(self, value):
        from xotl.tools.fp.option import Wrong

        try:
            return self.inner(value)
        except Exception as error:
            return Wrong(error)


class MultiCheck(predicate):
    """Return a wrong value only when all inner predicates fails.

    Haskell: guards (pp. 132)

    """

    __slots__ = ()

    def __new__(cls, *args):
        inner = tuple(predicate(arg) for arg in args)
        self = super().__new__(cls)
        self.inner = inner
        return self

    def __call__(self, value):
        from xotl.tools.fp.option import Just, Wrong, none

        predicates = self.inner
        i, res = 0, none
        while isinstance(res, Wrong) and i < len(predicates):
            res = predicates[i](value)
            i += 1
        return res.inner if isinstance(res, Just) and res.inner else res

    def __str__(self):
        aux = " OR ".join(str(c) for c in self.inner)
        return "combo({})".format(aux)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

import sys
from xotl.tools.values import file_coerce, positive_int_coerce as positive_int

# old params
from xotl.tools.params import ParamSchemeRow as row, ParamScheme as scheme


sample_scheme = scheme(
    row("stream", 0, -1, "output", default=sys.stdout, coerce=file_coerce),
    row("indent", 0, 1, default=1, coerce=positive_int),
    row("width", 0, 1, 2, "max_width", default=79, coerce=positive_int),
    row("newline", default="\n", coerce=(str,)),
)

del file_coerce, positive_int


def test_basic_params():
    def get_values(*args, **kwargs):
        return sample_scheme(args, kwargs, strict=False)

    def foobar(**kwargs):
        res = sample_scheme.defaults
        res.update(kwargs)
        return res

    one, two = get_values(4, 80), foobar(indent=4, width=80)
    assert one == two, "\n{} != \n{}".format(one, two)
    one, two = get_values(2), foobar(indent=2)
    assert one == two, "\n{} != \n{}".format(one, two)
    one = get_values(80, indent=4, extra="I'm OK!")
    two = foobar(width=80, indent=4, extra="I'm OK!")
    assert one == two, "\n{} != \n{}".format(one, two)
    one, two = get_values(width=80), foobar(width=80)
    assert one == two, "\n{} != \n{}".format(one, two)
    one = get_values(sys.stderr, 4, 80)
    two = foobar(indent=4, width=80, stream=sys.stderr)
    assert one == two, "\n{} != \n{}".format(one, two)
    one = get_values(4, sys.stderr, newline="\n\r")
    two = foobar(indent=4, stream=sys.stderr, newline="\n\r")
    assert one == two, "\n{} != \n{}".format(one, two)
    one = get_values(4, output=sys.stderr)
    two = foobar(indent=4, stream=sys.stderr)
    assert one == two, "\n{} != \n{}".format(one, two)
    one, two = get_values(4, max_width=80), foobar(indent=4, width=80)
    assert one == two, "\n{} != \n{}".format(one, two)


def test_param_errors():
    def get_values(*args, **kwargs):
        return sample_scheme(args, kwargs)

    def error_repr(error):
        return "{}()".format(type(error).__name__, error)

    msg = 'Must raised "{}", \n\tnot {}'

    try:
        get_values(sys.stderr, 4, output=sys.stderr)
        assert False, "Should raise TypeError"
    except TypeError:
        pass
    except BaseException as error:
        assert False, msg.format(TypeError.__name__, error_repr(error))
    try:
        get_values(4, -79)
        assert False, "Should raise TypeError"
    except TypeError:
        pass
    except BaseException as error:
        assert False, msg.format(TypeError.__name__, error_repr(error))
    try:
        get_values(80, indent=4, extra="I'm not OK!")
        assert False, "Should raise TypeError"
    except TypeError:
        pass
    except BaseException as error:
        assert False, msg.format(TypeError.__name__, error_repr(error))


def test_pop_keyword_values():
    from xotl.tools.symbols import Unset
    from xotl.tools.params import pop_keyword_values as popkw, Undefined

    kwds = dict(default=None, values=[1, 2, 3], collector=sum)
    names = (("func", "collector"), "default")

    assert popkw(dict(kwds), "values", *names) == [[1, 2, 3], sum, None]

    try:
        assert popkw(dict(kwds), *names) == "whatever"
    except TypeError:
        assert True

    try:
        assert popkw(dict(kwds), *names, ignore_error=True) == [sum, None]
    except TypeError:
        assert False

    test = [Undefined, [1, 2, 3], sum, None]
    assert popkw(dict(kwds), "x", "values", *names) == test

    test = [None, [1, 2, 3], sum, None]
    assert popkw(dict(kwds), "x", "values", *names, default=None) == test

    test = [Unset, [1, 2, 3], sum, None]
    defaults = dict(x=Unset)
    assert popkw(dict(kwds), "x", "values", *names, defaults=defaults) == test



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

"""Extends the standard `calendar` module.

xotl.tools extensions:

- Settle constants for all months.

Original documentation:

"""

from calendar import *  # noqa
import calendar as _stdlib  # noqa

__doc__ += _stdlib.__doc__

try:
    __all__ = list(_stdlib.__all__)
except AttributeError:
    pass


(
    January,
    February,
    March,
    April,
    May,
    June,
    July,
    August,
    September,
    October,
    November,
    December,
) = range(1, 13)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#


def test_csv():
    from xotl.tools.future.csv import parse, DefaultDialect

    data = [
        "A,B,C,D",
        "1,1.2,1.3,Spanish",
        "2,2.2,2.3,Español",
        '''"One, comma","""double quotes""",I'm a single quote,Inglés''',
    ]

    def forge(cell):
        try:
            return int(cell)
        except ValueError:
            try:
                return float(cell)
            except ValueError:
                return cell

    matrix = [[forge(cell) for cell in row] for row in parse(data)]

    sum_int, sum_float = 0, 0.0
    count_int, count_float, count_text = 0, 0, 0
    for row in matrix:
        for cell in row:
            if isinstance(cell, int):
                count_int += 1
                sum_int += cell
            elif isinstance(cell, float):
                count_float += 1
                sum_float += cell
            else:
                assert isinstance(cell, str)
                count_text += 1

    assert count_int == 2
    assert sum_int == 3
    assert sum_float == 7.0
    assert count_float == 4
    assert count_text == 10



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

import pytest
from hypothesis import strategies as s, given


def test_first_n_no_filling():
    from xotl.tools.future.itertools import first_n

    with pytest.raises(StopIteration):
        next(first_n((), 1))


def test_first_n_filling_by_cycling():
    from xotl.tools.future.itertools import first_n

    assert list(first_n((), 10, range(5))) == [0, 1, 2, 3, 4] * 2


def test_first_n_repeat_filling_by_repeating():
    from xotl.tools.future.itertools import first_n
    from itertools import repeat

    assert list(first_n((), 10, "0")) == list(repeat("0", 10))


def test_first_n_simple():
    from xotl.tools.future.itertools import first_n

    assert list(first_n(range(100), 10, 0)) == [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]


def test_slides():
    from xotl.tools.future.itertools import slides

    assert list(slides(range(1, 11))) == [(1, 2), (3, 4), (5, 6), (7, 8), (9, 10)]


def test_slides_filling():
    from xotl.tools.future.itertools import slides

    assert list(slides(range(1, 5), 3)) == [(1, 2, 3), (4, None, None)]


def test_slides_with_repeating_filling():
    from xotl.tools.future.itertools import slides

    aux = [(1, 2, 3), (4, 5, 6), (7, 8, 9), (10, None, None)]
    assert list(slides(range(1, 11), width=3, fill=None)) == aux


def test_slides_with_cycling_filling():
    from xotl.tools.future.itertools import slides

    aux = [(1, 2, 3, 4, 5), (6, 7, 8, 9, 10), (11, 1, 2, 1, 2)]
    assert list(slides(range(1, 12), width=5, fill=(1, 2))) == aux


def test_continuously_slides():
    from xotl.tools.future.itertools import continuously_slides

    aux = continuously_slides("maupasant", 3, "")
    trigrams = list("".join(x) for x in aux)
    assert "mau" in trigrams
    assert "aup" in trigrams
    assert "upa" in trigrams
    assert "pas" in trigrams
    assert "asa" in trigrams
    assert "san" in trigrams
    assert "ant" in trigrams
    assert len(trigrams) == 7


@s.composite
def keys(draw):
    return "k%d" % draw(s.integers(min_value=0, max_value=100))


@given(s.dictionaries(keys(), s.integers()), s.dictionaries(keys(), s.integers()))
def test_dict_update_new(d1, d2):
    from xotl.tools.future.itertools import dict_update_new

    d = dict(d1)
    dict_update_new(d1, d2)
    assert all(key in d1 for key in d2)
    assert all(d1[key] == d2[key] for key in d2 if key not in d)


@given(s.lists(s.integers(), max_size=30))
def test_delete_duplicates(l):
    from xotl.tools.future.itertools import delete_duplicates
    from xotl.tools.future.collections import Counter

    res = delete_duplicates(l)
    assert type(l) is type(res)  # noqa
    assert len(res) <= len(l)
    assert all(Counter(res)[item] == 1 for item in l)


@given(s.lists(s.integers(), max_size=30))
def test_delete_duplicates_with_key(l):
    from xotl.tools.future.itertools import delete_duplicates

    res = delete_duplicates(l, key=lambda x: x % 3)
    assert len(res) <= 3, "key yields 0, 1, or 2; thus res can contain at most 3 items"


def test_iter_delete_duplicates():
    from xotl.tools.future.itertools import iter_delete_duplicates

    assert list(iter_delete_duplicates("AAAaBBBA")) == ["A", "a", "B", "A"]
    assert list(iter_delete_duplicates("AAAaBBBA", key=lambda x: x.lower())) == [
        "A",
        "B",
        "A",
    ]

    assert list(iter_delete_duplicates("AAAaBBBA")) == ["A", "a", "B", "A"]
    assert list(iter_delete_duplicates("AAAaBBBA", key=lambda x: x.lower())) == [
        "A",
        "B",
        "A",
    ]


@given(
    s.lists(s.integers(), max_size=30),
    s.lists(s.integers(), max_size=30),
    s.lists(s.integers(), max_size=30),
)
def test_merge(l1, l2, l3):
    from xotl.tools.future.itertools import merge

    l1 = sorted(l1)
    l2 = sorted(l2)
    l3 = sorted(l3)
    # Accumulate and catch if yielding more than necessary
    iter_ = merge(l1, l2, l3)
    expected = sorted(l1 + l2 + l3)
    result = []
    for _ in range(len(expected)):
        result.append(next(iter_))
    with pytest.raises(StopIteration):
        last = next(iter_)  # noqa: There cannot be more items in the merge
    assert result == expected


@given(s.lists(s.integers(), max_size=30), s.lists(s.integers(), max_size=30))
def test_merge_by_key(l1, l2):
    from xotl.tools.future.itertools import merge

    l1 = [("l1-dummy", i) for i in sorted(l1)]
    l2 = [("l2-dummy", i) for i in sorted(l2)]
    # Accumulate and catch if yielding more than necessary
    iter_ = merge(l1, l2, key=lambda x: x[1])
    expected = sorted(l1 + l2, key=lambda x: x[1])
    result = []
    for _ in range(len(expected)):
        result.append(next(iter_))
    with pytest.raises(StopIteration):
        last = next(iter_)  # noqa: There cannot be more items in the merge
    assert result == expected


@given(s.lists(s.integers(), max_size=30), s.lists(s.integers(), max_size=30))
def test_merge_by_key_incomparable(l1, l2):
    class item:
        def __init__(self, x):
            self.item = x

    from xotl.tools.future.itertools import merge

    l1 = [item(i) for i in sorted(l1)]
    l2 = [item(i) for i in sorted(l2)]
    # Accumulate and catch if yielding more than necessary
    iter_ = merge(l1, l2, key=lambda x: x.item)
    expected = sorted(l1 + l2, key=lambda x: x.item)
    result = []
    for _ in range(len(expected)):
        result.append(next(iter_))
    with pytest.raises(StopIteration):
        last = next(iter_)  # noqa: There cannot be more items in the merge
    assert result == expected



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

"""A simple `main`:func: entry point for CLI based applications.

This module provides an example of how to use `xotl.tools.cli`:mod: to create a
CLI application.

"""


def main(default=None):
    """Execute a command.

    It can be given as the first program argument or it's the `default`
    command is defined.

    """
    import sys
    from xotl.tools.cli import Command, HELP_NAME

    args = sys.argv[1:]
    if args and not args[0].startswith("-"):
        cmd_name = args[0]
        args = args[1:]
    else:
        cmd_name = default or Command.get_setting("default_command", None) or HELP_NAME
    cmds = Command.registry
    cmd = cmds.get(cmd_name)
    if not cmd:
        print('Command "%s" not found!\n' % cmd_name)
        cmd = cmds.get(HELP_NAME)
        args = []
    sys.exit(cmd().run(args))


if __name__ == "__main__":
    main()
